import re
import warnings
from typing import Union, List, Optional

import pint.errors
from ontolutils import namespaces, urirefs
from ontolutils.utils.qudt_units import qudt_lookup
from pydantic import HttpUrl, field_validator, Field, ConfigDict
from pydantic import ValidationError
from pydantic_core import InitErrorDetails

from ssnolib import config
from ssnolib.dcat import Dataset
from ssnolib.skos import Concept
from .unit_utils import _parse_unit, _format_unit


@namespaces(ssno="https://matthiasprobst.github.io/ssno#",
            dcat="http://www.w3.org/ns/dcat#")
@urirefs(StandardName='ssno:StandardName',
         standardName='ssno:standardName',
         unit='ssno:unit',
         description='ssno:description',
         standardNameTable='ssno:standardNameTable')
class StandardName(Concept):
    """Implementation of ssno:StandardName"""
    model_config = ConfigDict(
        populate_by_name=True,
        # Whether an aliased field may be populated by its name as given by the model attribute, as well as the alias.
        extra="ignore",  # ignore any extra fields that are not defined in the model
    )

    standardName: str = Field(alias="standard_name")
    unit: Union[str, HttpUrl]  # required!
    description: Union[str, List[str]] = None  # ssno:description
    standardNameTable: Union[str, Dataset, "StandardNameTable"] = Field(default=None, alias="standard_name_table")
    alias: Optional[Union["StandardName", HttpUrl, str]] = Field(default=None)

    def __getattr__(self, item):
        for field, meta in self.model_fields.items():
            if meta.alias == item:
                return getattr(self, field)
        return super().__getattr__(item)

    def __setattr__(self, key, value):
        for field, meta in self.model_fields.items():
            if meta.alias == key:
                return setattr(self, field, value)
        return super().__setattr__(key, value)

    def __str__(self) -> str:
        if self.standardName is None:
            return ''
        return self.standardName

    @field_validator("alias", mode='before')
    @classmethod
    def _alias(cls, alias: Optional["StandardName"]) -> "StandardName":
        """Parse the alias and return the alias as StandardName."""
        if alias is None:
            return None
        if isinstance(alias, str):
            if alias.startswith('_:'):
                return str(alias)
            return HttpUrl(alias)
        elif not isinstance(alias, StandardName):
            raise TypeError(f"Expected a StandardName, got {type(alias)}")
        return alias

    @field_validator("standardNameTable", mode='before')
    @classmethod
    def _parse_standard_name_table(cls, standardNameTable: Union[Dataset, str]) -> Dataset:
        """Parse the standard_name_table and return the standard_name_table as Dataset."""
        from .standard_name_table import StandardNameTable
        if isinstance(standardNameTable, Dataset):
            return standardNameTable
        elif isinstance(standardNameTable, StandardNameTable):
            return standardNameTable
        elif isinstance(standardNameTable, str):
            assert standardNameTable.startswith('http'), f"Expected a URL, got {standardNameTable}"
            return str(HttpUrl(standardNameTable))
            # from .standard_name_table import StandardNameTable
            # return StandardNameTable(id=standardNameTable)
        raise TypeError(f"Expected a Dataset, got {type(standardNameTable)}")

    @field_validator("unit", mode='before')
    @classmethod
    def _parse_unit(cls, unit: Union[HttpUrl, str], cfg) -> str:
        """Parse the unit and return the unit as string."""
        if unit is None or unit in ('', '1', '-', 1):
            return str(qudt_lookup['dimensionless'])
        if isinstance(unit, str):
            if unit.startswith('http'):
                return str(HttpUrl(unit))
            err = False
            try:
                return str(qudt_lookup[unit.strip()])
            except KeyError:
                err = True

            try:
                _unit = _format_unit(str(_parse_unit(unit.strip())))
                return str(qudt_lookup[_unit])
            except (AttributeError, KeyError, pint.errors.DimensionalityError) as _:
                err = True

            if err:
                if config.raise_error_on_unparsable_unit:
                    err = InitErrorDetails(
                        type="value_error",
                        loc=("unit",),
                        input=unit,
                        ctx={"error": f'your_message Unable to parse: "{unit}" of standard name '
                                      f'"{cfg.data["standardName"]}"', }
                    )
                    raise ValidationError.from_exception_data(title=cls.__name__, line_errors=[err, ])
                else:
                    warnings.warn(f'Could not parse unit: "{unit}".', UserWarning)
            return str(unit)
        return str(HttpUrl(unit))

    @field_validator("standardName", mode='before')
    @classmethod
    def _parse_standard_name(cls, standardName: str) -> str:
        """Parse the standardName and return the standardName as string."""
        if not re.match(config.standard_name_core_pattern, standardName):
            raise ValueError(f"Invalid standard name '{standardName}' according to the core pattern "
                             f"'{config.standard_name_core_pattern}'.")
        return str(standardName)


@namespaces(ssno="https://matthiasprobst.github.io/ssno#")
@urirefs(ScalarStandardName='ssno:ScalarStandardName')
class ScalarStandardName(StandardName):
    pass


@namespaces(ssno="https://matthiasprobst.github.io/ssno#")
@urirefs(VectorStandardName='ssno:VectorStandardName')
class VectorStandardName(StandardName):
    pass



import pathlib
import unittest

__this_dir__ = pathlib.Path(__file__).parent

from ssnolib.sparql_utils import build_simple_sparql_query, WHERE
import rdflib


class TestSparqlUtils(unittest.TestCase):

    def test_build_sparql_query(self):
        sparql_query = build_simple_sparql_query(dict(
            owl="http://www.w3.org/2002/07/owl#",
            rdfs="http://www.w3.org/2000/01/rdf-schema#"
        ),
            wheres=[WHERE(s="?id", p="a", o="ssno:StandardNameTable")]
        )
        expected_sparql_string = """PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?id
WHERE {?id a ssno:StandardNameTable .
}"""
        self.assertEqual(expected_sparql_string, sparql_query.query_string)

    def test_perform_simple_query(self):
        test_data = """
{
    "@context": {
        "hdf5": "http://purl.allotrope.org/ontologies/hdf5/1.8#",
        "m4i": "http://w3id.org/nfdi4ing/metadata4ing#",
        "ssno": "https://matthiasprobst.github.io/ssno#"
    },
    "@graph": [
        {
            "@id": "_:N139",
            "@type": "hdf5:File",
            "hdf5:rootGroup": {
                "@id": "_:N138",
                "@type": [
                    "hdf5:Group",
                    "https://www.wikidata.org/wiki/Q1058834"
                ],
                "hdf5:attribute": [
                    {
                        "@id": "_:N140",
                        "@type": "hdf5:Attribute",
                        "hdf5:name": "fan_type",
                        "hdf5:value": "centrifugal"
                    },
                    {
                        "@id": "_:N141",
                        "@type": "hdf5:Attribute",
                        "hdf5:name": "manufacturing_method",
                        "hdf5:value": "rapid_prototyping"
                    }
                ]
            }
        }
    ]
}
    """
        spql = build_simple_sparql_query(
            prefixes={
                "ssno": "https://matthiasprobst.github.io/ssno#",
                "hdf5": "http://purl.allotrope.org/ontologies/hdf5/1.8#"
            },
            wheres=[
                WHERE("?id", "a", "hdf5:Attribute"),
                WHERE("?id", "hdf5:value", "\"centrifugal\"")
            ]
        )
        g = rdflib.Graph()
        g.parse(data=test_data,
                format='json-ld',
                context={
                    "ssno": "https://matthiasprobst.github.io/ssno#",
                    "hdf5": "http://purl.allotrope.org/ontologies/hdf5/1.8#"
                })
        res = spql.query(g)
        print(res)



import unittest

import pydantic

from ssnolib.hdf5 import File, Dataset, Group


class TestHDF5(unittest.TestCase):

    def testHDF5File(self):
        root_group = Group(name="/")
        self.assertEqual(root_group.name, "/")
        file = File(rootGroup=root_group)
        self.assertEqual(file.rootGroup.name, "/")

        group = Group(name="/grp")
        with self.assertRaises(pydantic.ValidationError):
            File(rootGroup=group)

    def testDataset(self):
        with self.assertRaises(pydantic.ValidationError):
            Dataset(name='Dataset1')
        ds1 = Dataset(name='/Dataset1')
        self.assertEqual(ds1.name, '/Dataset1')

    def testGroup(self):
        with self.assertRaises(pydantic.ValidationError):
            Group(name='Group1')
        grp1 = Group(name='/Group1')



import pathlib
from typing import Union

import yaml

from ssnolib import StandardName


def read_standard_names_from_yaml(yaml_filename: Union[str, pathlib.Path]):
    yaml_filename = pathlib.Path(yaml_filename)
    if not yaml_filename.exists():
        raise FileNotFoundError(f"File {yaml_filename} does not exist")

    with open(yaml_filename, 'r') as f:
        data = yaml.safe_load(f)

    standard_names = []
    for k, v in data.items():
        standard_names.append(StandardName(standard_name=k, **v))

    return standard_names



import json
import unittest

from ssnolib import StandardName
from ssnolib.pimsii import Property
from ssnolib.pimsii.variable import Variable


class TestPIMSII(unittest.TestCase):

    def testVariable(self):
        variable = Variable(
            id="_:b1",
            label="my variable"
        )
        self.assertDictEqual(
            {
                "@context": {
                    "owl": "http://www.w3.org/2002/07/owl#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "pims": "http://www.molmod.info/semantics/pims-ii.ttl#",
                    "m4i": "http://w3id.org/nfdi4ing/metadata4ing#"
                },
                "@type": "pims:Variable",
                "rdfs:label": "my variable",
                "@id": "_:b1"
            },
            json.loads(variable.model_dump_jsonld())
        )

    def testProperty(self):
        prop = Property(
            id="_:b1",
            label="my property",
            hasValue=5.4,
            hasStandardName=StandardName(
                id="_:b2",
                standardName='x_velocity',
                unit='m/s')
        )
        self.assertDictEqual(
            {
                "@context": {
                    "owl": "http://www.w3.org/2002/07/owl#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "pims": "http://www.molmod.info/semantics/pims-ii.ttl#",
                    "m4i": "http://w3id.org/nfdi4ing/metadata4ing#",
                    "ssno": "https://matthiasprobst.github.io/ssno#",
                    "skos": "http://www.w3.org/2004/02/skos/core#",
                    "dcat": "http://www.w3.org/ns/dcat#"
                },
                "@type": "pims:Property",
                "rdfs:label": "my property",
                "m4i:hasValue": 5.4,
                "ssno:hasStandardName": {
                    "@type": "ssno:StandardName",
                    "ssno:standardName": "x_velocity",
                    "ssno:unit": "http://qudt.org/vocab/unit/M-PER-SEC",
                    "@id": "_:b2"
                },
                "@id": "_:b1"
            },
            json.loads(prop.model_dump_jsonld())
        )



import pathlib
import unittest

import ontolutils
import pydantic
from ontolutils import QUDT_UNIT
from ontolutils.utils.qudt_units import parse_unit
from pydantic import ValidationError

import ssnolib
from ssnolib import StandardName, StandardNameTable

__this_dir__ = pathlib.Path(__file__).parent

CACHE_DIR = ssnolib.utils.get_cache_dir()


class TestSSNOStandardName(unittest.TestCase):

    def test_standard_name_and_table(self):
        snt = StandardNameTable(title="my snt")
        sn = StandardName(
            standard_name='x_velocity',
            unit="m/s",
            description='x component of velocity',
            standardNameTable=snt
        )
        self.assertEqual(sn.standardNameTable.title, "my snt")

    def test_standard_name_alias(self):
        sn = StandardName(standard_name='x_velocity',
                          description='x component of velocity',
                          unit='m/s')
        sn_alias = StandardName(standardName='x_velocity',
                                description='x component of velocity',
                                unit='m/s',
                                alias=sn)
        self.assertEqual(sn.standard_name, 'x_velocity')
        self.assertEqual(sn.description, 'x component of velocity')
        self.assertEqual(sn.unit, sn_alias.unit)

        sn_alias = StandardName(standardName='x_velocity',
                                description='x component of velocity',
                                unit='m/s',
                                alias=sn.id)
        self.assertEqual(sn_alias.alias, sn.id)
        self.assertEqual(sn.standardName, sn_alias.standardName)
        self.assertEqual(sn.unit, sn_alias.unit)
        self.assertEqual(sn.description, sn_alias.description)

    def test_units(self):
        sn = StandardName(
            standard_name='x_velocity',
            description='x component of velocity',
            unit=''
        )
        self.assertEqual(sn.unit, str(QUDT_UNIT.UNITLESS))
        sn = StandardName(
            standard_name='x_velocity',
            description='x component of velocity',
            unit='1'
        )
        self.assertEqual(sn.unit, str(QUDT_UNIT.UNITLESS))

    def test_instantiating_standard_name_missing_fields(self):
        with self.assertRaises(pydantic.ValidationError):
            sn = StandardName()
        with self.assertRaises(pydantic.ValidationError):
            sn = StandardName(standardName='x_velocity', )
        with self.assertRaises(pydantic.ValidationError):
            sn = StandardName(standardName='x_velocity',
                              description='x component of velocity')

    def test_instantiating_standard_name(self):
        snt = StandardNameTable()
        sn = StandardName(standardName='x_velocity',
                          description='x component of velocity',
                          unit=QUDT_UNIT.M_PER_SEC,
                          standardNameTable=snt)
        self.assertIsInstance(sn, ontolutils.Thing)
        self.assertIsInstance(sn, StandardName)
        self.assertEqual(sn.standardName, 'x_velocity')
        self.assertEqual(sn.standard_name, 'x_velocity')
        self.assertEqual(sn.description, 'x component of velocity')
        self.assertEqual(sn.unit, str(parse_unit('m s-1')))
        self.assertEqual(sn.unit, str(parse_unit('m s-1')))
        self.assertIsInstance(sn.standard_name_table, StandardNameTable)
        self.assertEqual(snt.id, sn.standardNameTable.id)

        sn = StandardName(standardName='x_velocity',
                          description='x component of velocity',
                          unit=None)
        self.assertEqual(str(sn.unit), "http://qudt.org/vocab/unit/UNITLESS")

        with self.assertRaises(pydantic.ValidationError):
            StandardName(standardName='x_velocity',
                         description='x component of velocity',
                         unit="213nlsfh8os")

        try:
            StandardName(standardName='x_velocity',
                         description='x component of velocity',
                         unit="213nlsfh8os")
        except pydantic.ValidationError as e:
            self.assertEqual(e.errors()[0]['loc'], ('unit', 'unit'))
            self.assertEqual(e.errors()[0]['msg'],
                             'Value error, your_message Unable to parse: "213nlsfh8os" of standard name "x_velocity"')

        sn = StandardName(standardName='x_velocity',
                          description='x component of velocity',
                          unit=QUDT_UNIT.M_PER_SEC)
        self.assertEqual(str(sn.unit), 'http://qudt.org/vocab/unit/M-PER-SEC')

    def test_instantiating_standard_name_using_aliases(self):
        snt = StandardNameTable()
        sn = StandardName(standard_name='x_velocity',
                          description='x component of velocity',
                          unit=QUDT_UNIT.M_PER_SEC,
                          standard_name_table=snt)
        self.assertIsInstance(sn, ontolutils.Thing)
        self.assertIsInstance(sn, StandardName)
        self.assertEqual(sn.standardName, 'x_velocity')
        self.assertEqual(sn.standard_name, 'x_velocity')
        self.assertEqual(sn.description, 'x component of velocity')
        self.assertEqual(sn.unit, str(parse_unit('m s-1')))
        self.assertEqual(sn.unit, str(parse_unit('m s-1')))
        self.assertIsInstance(sn.standard_name_table, StandardNameTable)
        self.assertEqual(snt.id, sn.standard_name_table.id)

    def test_instantiating_standard_name_with_snt(self):
        sn = StandardName(standard_name='x_velocity',
                          unit=QUDT_UNIT.M_PER_SEC,
                          standard_name_table="https://doi.org/10.5281/zenodo.10428817")
        self.assertEqual(sn.standardNameTable, "https://doi.org/10.5281/zenodo.10428817")

        with self.assertRaises(pydantic.ValidationError):
            StandardName(standard_name='x_velocity',
                         unit=QUDT_UNIT.M_PER_SEC,
                         standard_name_table="123")

    def test_standard_name_jsonld(self):
        sn = StandardName(standard_name='x_velocity',
                          description='x component of velocity',
                          unit='m s-1')
        self.assertEqual(sn.unit, str(parse_unit('m s-1')))

        with open('sn.jsonld', 'w') as f:
            f.write(sn.model_dump_jsonld())

        sn_loaded = ontolutils.query(StandardName, source='sn.jsonld')
        self.assertEqual(len(sn_loaded), 1)
        self.assertEqual(sn_loaded[0].standardName, 'x_velocity')
        self.assertEqual(sn_loaded[0].description, 'x component of velocity')
        self.assertEqual(sn_loaded[0].unit, str(parse_unit('m s-1')))

        sn_loaded = StandardName.from_jsonld(data=sn.model_dump_jsonld())
        self.assertEqual(len(sn_loaded), 1)
        self.assertEqual(sn_loaded[0].standardName, 'x_velocity')
        self.assertEqual(sn_loaded[0].description, 'x component of velocity')
        self.assertEqual(sn_loaded[0].unit, str(parse_unit('m s-1')))

        pathlib.Path('sn.jsonld').unlink(missing_ok=True)

    def test_invalid_standard_names(self):
        # Wrong type for description:
        with self.assertRaises(pydantic.ValidationError):
            ssnolib.StandardName(
                standard_name='x_velocity',
                unit='m/s',
                description=123
            )

        # Incorrect standard name that does not match the basic pattern:
        with self.assertRaises(pydantic.ValidationError):
            ssnolib.StandardName(
                standard_name='X_velocity_',
                unit='m/s',
                description="invalid standard name"
            )

        # Cannot parse unit
        with self.assertRaises(ValidationError):
            ssnolib.StandardName(
                standard_name='x_velocity',
                unit='meterprosec',
                description="invalid standard name"
            )

        # Missing field(s)
        with self.assertRaises(pydantic.ValidationError):
            ssnolib.StandardName(
                standard_name='x_velocity',
                description="invalid standard name"
            )

    def test_standard_name_with_table(self):
        snt = StandardNameTable(identifier='https://doi.org/10.5281/zenodo.10428817')

        xvel = StandardName(
            standard_name="x_velocity",
            description="x component of velocity",
            unit="m/s",
            standardNameTable=snt
        )
        self.assertEqual(xvel.__str__(), 'x_velocity')
        self.assertEqual(xvel.standardNameTable, snt)

    def test_standard_name_with_difficult_unit(self):
        sn = StandardName(standardName="velocity", description="velocity", unit="m2/s2")
        self.assertEqual(str(sn.unit), 'http://qudt.org/vocab/unit/M2-PER-SEC2')



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class SSNO(DefinedNamespace):
    # uri = "https://matthiasprobst.github.io/ssno#"
    # Generated with ssnolib
    Character: URIRef  # ['Charakter']
    Qualification: URIRef  # ['Qualification']
    ScalarStandardName: URIRef  # ['Scalar Standard Name']
    StandardName: URIRef  # ['Standard Name']
    StandardNameModification: URIRef  # ['Standard Name Modification']
    StandardNameTable: URIRef  # ['Standard NameTable']
    Transformation: URIRef  # ['Transformation']
    VectorQualification: URIRef  # ['Vector Qualification']
    VectorStandardName: URIRef  # ['Vector Standard Name']
    after: URIRef  # ['after']
    alias: URIRef  # ['alias']
    associatedWith: URIRef  # ['associated with']
    before: URIRef  # ['before']
    dataset: URIRef  # ['dataset']
    hasCharacter: URIRef  # ['has character']
    hasModifier: URIRef  # ['has modifier']
    hasStandardName: URIRef  # ['has standard name']
    hasValidValues: URIRef  # ['has valid values']
    isStandardNameOf: URIRef  # ['is standard name of']
    positioned: URIRef  # ['positioned']
    standardNameTable: URIRef  # ['standard name table']
    standardNameTableUsedBy: URIRef  # ['standard name table used by']
    standardNames: URIRef  # ['standard names']
    unit: URIRef  # ['unit']
    usesStandardNameTable: URIRef  # ['uses standard name table']
    altersUnit: URIRef  # ['alters unit']
    ancillaryVariables: URIRef  # ['ancillary variables']
    character: URIRef  # ['character']
    description: URIRef  # ['description']
    hasPreposition: URIRef  # ['has preposition']
    longName: URIRef  # ['long name']
    standardName: URIRef  # ['standard name']
    standardNameCharacter: URIRef  # ['standard name character']
    AnyStandardName: URIRef  # ['Any Standard Name']
    AnyUnit: URIRef  # ['any unit']

    _NS = Namespace("https://matthiasprobst.github.io/ssno#")


setattr(SSNO, "Charakter", SSNO.Character)
setattr(SSNO, "Qualification", SSNO.Qualification)
setattr(SSNO, "Scalar_Standard_Name", SSNO.ScalarStandardName)
setattr(SSNO, "Standard_Name", SSNO.StandardName)
setattr(SSNO, "Standard_Name_Modification", SSNO.StandardNameModification)
setattr(SSNO, "Standard_NameTable", SSNO.StandardNameTable)
setattr(SSNO, "Transformation", SSNO.Transformation)
setattr(SSNO, "Vector_Qualification", SSNO.VectorQualification)
setattr(SSNO, "Vector_Standard_Name", SSNO.VectorStandardName)
setattr(SSNO, "after", SSNO.after)
setattr(SSNO, "alias", SSNO.alias)
setattr(SSNO, "associated_with", SSNO.associatedWith)
setattr(SSNO, "before", SSNO.before)
setattr(SSNO, "dataset", SSNO.dataset)
setattr(SSNO, "has_character", SSNO.hasCharacter)
setattr(SSNO, "has_modifier", SSNO.hasModifier)
setattr(SSNO, "has_standard_name", SSNO.hasStandardName)
setattr(SSNO, "has_valid_values", SSNO.hasValidValues)
setattr(SSNO, "is_standard_name_of", SSNO.isStandardNameOf)
setattr(SSNO, "positioned", SSNO.positioned)
setattr(SSNO, "standard_name_table", SSNO.standardNameTable)
setattr(SSNO, "standard_name_table_used_by", SSNO.standardNameTableUsedBy)
setattr(SSNO, "standard_names", SSNO.standardNames)
setattr(SSNO, "unit", SSNO.unit)
setattr(SSNO, "uses_standard_name_table", SSNO.usesStandardNameTable)
setattr(SSNO, "alters_unit", SSNO.altersUnit)
setattr(SSNO, "ancillary_variables", SSNO.ancillaryVariables)
setattr(SSNO, "character", SSNO.character)
setattr(SSNO, "description", SSNO.description)
setattr(SSNO, "has_preposition", SSNO.hasPreposition)
setattr(SSNO, "long_name", SSNO.longName)
setattr(SSNO, "standard_name", SSNO.standardName)
setattr(SSNO, "standard_name_character", SSNO.standardNameCharacter)
setattr(SSNO, "Any_Standard_Name", SSNO.AnyStandardName)
setattr(SSNO, "any_unit", SSNO.AnyUnit)


"""Implementation of dcat:Role
"""

from ontolutils import urirefs, namespaces

from ..skos import Concept


@namespaces(dcat="http://www.w3.org/ns/dcat#")
@urirefs(Role='dcat:Role')
class Role(Concept):
    """Pydantic implementation of dcat:Role"""



from dataclasses import dataclass
from typing import Dict, Optional, List

import rdflib


@dataclass
class WHERE:
    s: str
    p: str
    o: str
    return_variables: Optional[List[str]] = None
    is_optional: Optional[bool] = False

    def get_variables(self):
        if self.return_variables:
            return self.return_variables
        return_variables = []
        if self.s.startswith("?"):
            return_variables.append(self.s)
        if self.o.startswith("?"):
            return_variables.append(self.o)
        return return_variables

    def __post_init__(self):
        if self.s.startswith('http'):
            self.s = f"<{self.s}>"
        if self.p.startswith('http'):
            self.p = f"<{self.p}>"

    def __str__(self):
        if self.is_optional:
            return f"OPTIONAL {{{self.s} {self.p} {self.o}}} ."
        return f"{self.s} {self.p} {self.o} ."


@dataclass
class UNION:
    a: WHERE
    b: WHERE
    return_variables: Optional[List[str]] = None

    def get_variables(self):
        return_variables = []
        if self.a.return_variables:
            return_variables.extend(self.a.return_variables)
        if self.b.return_variables:
            return_variables.extend(self.b.return_variables)
        return return_variables

    def __str__(self):
        return f"{{ {self.a} }} UNION {{ {self.b} }}"


class SparqlQuery:
    def __init__(self, query_string, variables):
        self.query_string = query_string
        self.variables = variables

    def query(self, g: rdflib.Graph):
        results = g.query(self.query_string)
        rows = []
        for result in results:
            row = {}
            for var, res in zip(self.variables, result):
                row[var] = res
            rows.append(row)

        return rows


def build_simple_sparql_query(
        prefixes: Dict,
        wheres: List[WHERE]
) -> SparqlQuery:
    prefixes_str = "\n".join([f"PREFIX {k}: <{v}>" for k, v in prefixes.items()])
    return_variables = []
    where_strs = []
    for where in wheres:
        where_strs.append(str(where))

        if where.return_variables:
            for v in where.return_variables:
                return_variables.append(v)
        else:
            return_variables.extend(where.get_variables())

    set_of_ret_variables = set(return_variables)
    select_str = "SELECT " + ' '.join(set_of_ret_variables)
    where_str = "WHERE {" + '\n'.join(where_strs) + "\n}"
    sparql_str = f"{prefixes_str}\n{select_str}\n{where_str}"

    return SparqlQuery(sparql_str, [v[1:] for v in set_of_ret_variables])



from typing import Optional, Union

from ontolutils import Thing, namespaces, urirefs
from pydantic import Field


@namespaces(pims="http://www.molmod.info/semantics/pims-ii.ttl#",
            m4i="http://w3id.org/nfdi4ing/metadata4ing#")
@urirefs(Variable='pims:Variable',
         hasVariableDescription='m4i:hasVariableDescription',
         hasSymbol='m4i:hasSymbol',
         hasValue='m4i:hasValue',
         )
class Variable(Thing):
    hasVariableDescription: Optional[str] = Field(default=None, alias="has_variable_description")
    hasSymbol: Optional[str] = Field(default=None, alias="has_symbol")
    hasValue: Optional[Union[int, float]] = Field(default=None, alias="has_value")



from .standard_name import StandardName, ScalarStandardName, VectorStandardName
from .standard_name_table import StandardNameTable, Qualification, VectorQualification, Transformation, Character, \
    parse_table, AgentRole, StandardNameModification

__all__ = ('StandardNameTable',
           'Qualification',
           'VectorQualification',
           'ScalarStandardName',
           'VectorStandardName',
           'Transformation',
           'StandardName',
           'Character',
           'AgentRole',
           'StandardNameModification'
           )



try:
    from h5rdmtoolbox.wrapper.accessor import Accessor, register_accessor
    from h5rdmtoolbox.wrapper.core import Group
except ImportError:
    raise ImportError("h5rdmtoolbox is required for this function.")


@register_accessor("ssno", "group")
class SSNOAccessor(Accessor):
    """Accessor to await selected data to be converted to a new units"""

    def enrich_hdf(self,
                   standard_name_attribute="standard_name",
                   standard_name_table_attribute="standard_name_table") -> Group:
        """Add RDF information to an HDF5 file which has standard name attributes, i.e.
        datasets with a 'standard_name' attribute and the root group with a 'standard_name_table' attribute.

        Parameters
        ----------
        h5 : h5py.Group or str
            The root group of the HDF5 file or the filename of the HDF5 file.
        """
        h5 = self._obj  # root group

        # iteratively walk through the HDF5 file and add RDF information
        all_ds_with_sn = h5.find(flt={standard_name_attribute: {'$exists': True}}, objfilter='dataset')
        for ds in all_ds_with_sn:
            sn = ds.attrs[standard_name_attribute]
            h5[ds.name].rdf.predicate[
                standard_name_attribute] = "https://matthiasprobst.github.io/ssno#hasStandardName"

        snt_attr_val = h5.attrs.get(standard_name_table_attribute, None)
        if snt_attr_val is None:
            raise ValueError(f"Root group must have attribute '{standard_name_table_attribute}'")
        if not isinstance(snt_attr_val, str):
            raise ValueError(f"Attribute '{standard_name_table_attribute}' must be a string")
        if not snt_attr_val.startswith("http"):
            raise ValueError(f"Attribute '{standard_name_table_attribute}' must be a valid URI")
        h5.rdf.predicate[
            standard_name_table_attribute] = "https://matthiasprobst.github.io/ssno#hasStandardNameTable"
        h5.rdf.object[standard_name_table_attribute] = "https://matthiasprobst.github.io/ssno#StandardNameTable"
        return h5



import unittest

from ssnolib import StandardName
from ssnolib.m4i import TextVariable, NumericalVariable


class TestM4i(unittest.TestCase):

    def testTextVariable(self):
        text_variable = TextVariable(
            hasStringValue='String value',
            hasVariableDescription='Variable description'
        )
        self.assertEqual(text_variable.hasStringValue, 'String value')
        self.assertEqual(text_variable.hasVariableDescription, 'Variable description')

    def testNumericalVariableWithoutStandardName(self):
        numerical_variable = NumericalVariable(
            hasUnit='Unit',
            hasNumericalValue=1.0,
            hasMaximumValue=2.0,
            hasVariableDescription='Variable description')
        self.assertEqual(numerical_variable.hasUnit, 'Unit')
        self.assertEqual(numerical_variable.hasNumericalValue, 1.0)
        self.assertEqual(numerical_variable.hasMaximumValue, 2.0)
        self.assertEqual(numerical_variable.hasVariableDescription, 'Variable description')
        self.assertEqual(numerical_variable.hasStandardName, None)

    def testNumericalVariableWithStandardName(self):
        numerical_variable = NumericalVariable(
            hasUnit='mm/s',
            hasNumericalValue=1.0,
            hasMaximumValue=2.0,
            hasVariableDescription='Variable description',
            hasStandardName=StandardName(
                standardName='x_velocity',
                unit='m/s')
        )
        self.assertEqual(numerical_variable.hasUnit, 'mm/s')
        self.assertEqual(numerical_variable.hasNumericalValue, 1.0)
        self.assertEqual(numerical_variable.hasMaximumValue, 2.0)
        self.assertEqual(numerical_variable.hasVariableDescription, 'Variable description')
        self.assertEqual(numerical_variable.hasStandardName.standardName, 'x_velocity')
        self.assertEqual(numerical_variable.hasStandardName.unit, 'http://qudt.org/vocab/unit/M-PER-SEC')



import unittest

from rdflib import URIRef

from ssnolib.namespace import SSNO


class TestNamespace(unittest.TestCase):

    def test_namespace(self):
        self.assertIsInstance(SSNO.has_standard_names, URIRef)
        self.assertEqual(SSNO.has_standard_names, URIRef('https://matthiasprobst.github.io/ssno#has_standard_names'))
        self.assertEqual(SSNO.standardNames, URIRef('https://matthiasprobst.github.io/ssno#standardNames'))
        self.assertEqual(SSNO.StandardNames, URIRef('https://matthiasprobst.github.io/ssno#StandardNames'))



# """Implementation of classes
# - dcat:Resource
# - dcat:Distribution
# - dcat:Dataset
# """
# import pathlib
# import re
# import shutil
# from datetime import datetime
# from dateutil import parser
# from pydantic import HttpUrl, FileUrl, field_validator
# from typing import Union, List
#
# from .agent import Person, Organization
# from .core import Thing
# from .utils import download_file
#
#
# class Resource(Thing):
#     """Pyantic implementation of dcat:Resource
#
#     .. note::
#
#         More than the below parameters are possible but not explicitly defined here.
#
#
#
#     Parameters
#     ----------
#     title: str
#         Title of the resource (dcterms:title)
#     description: str = None
#         Description of the resource (dcterms:description)
#     creator: Union[Person, Organization] = None
#         Creator of the resource (dcterms:creator)
#     version: str = None
#         Version of the resource (dcat:version)
#     """
#     title: str  # dcterms:title
#     description: str = None  # dcterms:description
#     creator: Union[Person, Organization] = None  # dcterms:creator
#     version: str = None  # dcat:version
#
#     def _repr_html_(self):
#         """Returns the HTML representation of the class"""
#         return f"{self.__class__.__name__}({self.title})"
#
#
# class Distribution(Resource):
#     """Implementation of dcat:Distribution
#
#     .. note::
#         More than the below parameters are possible but not explicitly defined here.
#
#
#     Parameters
#     ----------
#     downloadURL: Union[HttpUrl, FileUrl]
#         Download URL of the distribution (dcat:downloadURL)
#     mediaType: HttpUrl = None
#         Media type of the distribution (dcat:mediaType).
#         Should be defined by the [IANA Media Types registry](https://www.iana.org/assignments/media-types/media-types.xhtml)
#     byteSize: int = None
#         Size of the distribution in bytes (dcat:byteSize)
#     """
#     downloadURL: Union[HttpUrl, FileUrl]  # dcat:downloadURL, e.g.
#     mediaType: HttpUrl = None  # dcat:mediaType
#     byteSize: int = None  # dcat:byteSize
#     keyword: List[str] = None  # dcat:keyword
#
#     def _repr_html_(self):
#         """Returns the HTML representation of the class"""
#         return f"{self.__class__.__name__}({self.downloadURL})"
#
#     def download(self,
#                  dest_filename: Union[str, pathlib.Path] = None,
#                  overwrite_existing: bool = False) -> pathlib.Path:
#         """Downloads the distribution"""
#
#         def _parse_file_url(furl):
#             """in windows, we might need to strip the leading slash"""
#             fname = pathlib.Path(furl)
#             if fname.exists():
#                 return fname
#             fname = pathlib.Path(self.downloadURL.path[1:])
#             if fname.exists():
#                 return fname
#             raise FileNotFoundError(f'File {self.downloadURL.path} does not exist')
#
#         if self.downloadURL.scheme == 'file':
#             if dest_filename is None:
#                 return _parse_file_url(self.downloadURL.path)
#             else:
#                 return shutil.copy(_parse_file_url(self.downloadURL.path), dest_filename)
#         return download_file(self.downloadURL,
#                              dest_filename,
#                              overwrite_existing=overwrite_existing)
#
#     @field_validator('mediaType', mode='before')
#     @classmethod
#     def _mediaType(cls, mediaType):
#         """should be a valid URI, like: https://www.iana.org/assignments/media-types/text/markdown"""
#         if isinstance(mediaType, str):
#             if mediaType.startswith('http'):
#                 return HttpUrl(mediaType)
#             elif mediaType.startswith('iana:'):
#                 return HttpUrl("https://www.iana.org/assignments/media-types/" + mediaType.split(":", 1)[-1])
#             elif re.match('[a-z].*/[a-z].*', mediaType):
#                 return HttpUrl("https://www.iana.org/assignments/media-types/" + mediaType)
#         return mediaType
#
#
# class Dataset(Resource):
#     """Pydantic implementation of dcat:Dataset
#
#     .. note::
#
#         More than the below parameters are possible but not explicitly defined here.
#
#
#
#     Parameters
#     ----------
#     title: str
#         Title of the resource (dcterms:title)
#     description: str = None
#         Description of the resource (dcterms:description)
#     creator: Union[Person, Organization] = None
#         Creator of the resource (dcterms:creator)
#     version: str = None
#         Version of the resource (dcat:version)
#     identifier: HttpUrl = None
#         Identifier of the resource (dcterms:identifier)
#     contact: Union[Person, Organization] = None
#         Contact person or organization of the resource (http://www.w3.org/ns/prov#Person)
#     distribution: List[Distribution] = None
#         Distribution of the resource (dcat:Distribution)
#     modified: datetime = None
#         Last modified date of the resource (dcterms:modified)
#     """
#     identifier: HttpUrl = None  # dcterms:identifier, see https://www.w3.org/TR/vocab-dcat-3/#ex-identifier
#     # http://www.w3.org/ns/prov#Person, see https://www.w3.org/TR/vocab-dcat-3/#ex-adms-identifier
#     contact: Union[Person, Organization] = None
#     distribution: Union[Distribution, List[Distribution]] = None  # dcat:Distribution
#     modified: datetime = None  # dcterms:modified
#
#     @field_validator('modified', mode='before')
#     @classmethod
#     def _modified(cls, modified):
#         """parse datetime"""
#         if isinstance(modified, str):
#             return parser.parse(modified)
#         return modified



import abc
import logging
import pathlib
import warnings
from typing import Dict, Union

from ssnolib.prov import Attribution
from . import ScalarStandardName, VectorStandardName

logger = logging.getLogger("ssnolib")


class TableReader(abc.ABC):
    """Abstract Standard Name Table Reader"""

    def __init__(self, filename: Union[str, pathlib.Path]):
        self.filename = pathlib.Path(filename)
        assert self.filename.exists(), f'{self.filename} does not exist'
        assert self.filename.is_file(), f'{self.filename} is not a file'

    @abc.abstractmethod
    def parse(self, **kwargs) -> Dict:
        """Parse the file"""


class XMLReader(TableReader):

    def parse(self, make_standard_names_lowercase: bool = False) -> Dict:
        """Parse the file"""
        try:
            import xmltodict
        except ImportError:
            raise ImportError('Package "xmltodict" is missing, but required to import from XML files.')

        with open(str(self.filename), 'r', encoding='utf-8') as file:
            my_xml = file.read()
        xmldict = xmltodict.parse(my_xml)
        _name = list(xmldict.keys())[0]

        xmldata = xmldict[_name]

        def _parse_standard_name(sndict):
            unit = sndict.get('canonical_units', '')
            if unit == '1':
                unit = ''
            elif unit is None:
                unit = ''
            description = sndict.get('description', '')
            if description is None:
                description = ''
            standard_name = sndict.get('@id')
            assert standard_name is not None, 'Expected key "@id" in the XML file.'
            assert unit is not None, 'Expected key "canonical_units" in the XML file.'
            assert description is not None, 'Expected key "description" in the XML file.'
            return dict(standardName=standard_name.lower() if make_standard_names_lowercase else standard_name,
                        unit=unit,
                        description=description)

        version = xmldata.get('version', None)
        if version is None:
            version = xmldata.get('version_number', None)

        # last_modified = xmldata.get('last_modified', None)

        contact = xmldata.get('contact', None)
        institution = xmldata.get('institution', None)
        if "@" in contact and institution is not None:
            # it is an email address
            from ssnolib.prov import Organization
            org = Organization(mbox=contact, name=institution)
            agent = Attribution(agent=org)
            if org.hasRorId:
                org.id = org.hasRorId
        else:
            agent = Attribution(agent=contact)

            # else cannot be parsed

        sn_data = xmldata.get('entry', None)
        if sn_data is None:
            raise KeyError('Expected key "entry" in the XML file.')
        data = {'standard_name': [_parse_standard_name(sn) for sn in sn_data],
                'version': version,
                # 'modified': last_modified,
                'qualifiedAttribution': agent}

        if 'title' not in xmldata:
            data['title'] = self.filename.stem

        sndata = data.pop('standard_name')
        for sn in sndata:
            if sn['description'] is None:
                name = sn['standard_name']
                warnings.warn(f'Description of "{name}" is None. Setting to empty string.', UserWarning)
                sn['description'] = ""
        data['standardNames'] = sndata
        return data


class YAMLReader(TableReader):

    def parse(self, **kwargs):
        try:
            import yaml
        except ImportError as e:
            raise ImportError('Package "pyyaml" is missing, but required to import from YAML files.') from e

        with open(self.filename, 'r') as f:
            data = yaml.safe_load(f)
        standardNames = data.get('standardNames', data.get("standard_names", {}))

        def _parse_standard_names(name, sndata: Dict):
            for ustr in ('unit', 'units', 'canonical_unit', 'canonicalUnits'):
                if ustr in sndata:
                    sndata['unit'] = sndata.pop(ustr)
                    break
            _data = {
                'standard_name': name,
                **sndata
            }
            is_vector = 'vector' in _data
            for k in list(_data.keys()):
                if k not in ('unit', 'description', 'standard_name'):
                    _data.pop(k)
            if is_vector:
                return VectorStandardName(**_data)
            return ScalarStandardName(**_data)

        qualifiedAttribution = data.get('creator', None)
        # make the orcid id the ID of the creator:
        if qualifiedAttribution:
            if isinstance(qualifiedAttribution, list):
                for ic, c in enumerate(qualifiedAttribution.copy()):
                    if c['orcid_id']:
                        qualifiedAttribution[ic]['id'] = c['orcid_id']
            else:
                if qualifiedAttribution['orcid_id']:
                    qualifiedAttribution['id'] = qualifiedAttribution['orcid_id']

        # parse qualifications
        qualification_data = data.get('qualifications', None)
        qualifications_dict = {}
        if qualification_data:
            construction = qualification_data.pop('construction', None)
            if construction is None:
                logger.error("No construction string is provided in the qualifications.")
            phrases = qualification_data.get('phrases', None)
            if phrases:
                from ssnolib.ssno.standard_name_table import Qualification
                qualifications = [Qualification(
                    name=q['name'],
                    description=q.get('description', None),
                    hasPreposition=q.get('hasPreposition', None),
                    hasValidValues=q.get('hasValidValues', None)
                ) for q in phrases]
                qualifications_dict = {q.get_full_name(): q for q in qualifications}
            else:
                logger.error("No phrases are provided in the qualifications.")

            if phrases and qualifications_dict:
                from ssnolib.utils import gpfqcs
                positions = gpfqcs(construction)
                for position, q_full_name in positions.items():
                    if position < 0:
                        if position == -1:
                            from ssnolib.namespace import SSNO
                            qualifications_dict[q_full_name].before = SSNO.AnyStandardName
                        else:
                            qualifications_dict[q_full_name].before = qualifications_dict[position + 1]
                    else:
                        if position == 1:
                            from ssnolib.namespace import SSNO
                            qualifications_dict[q_full_name.replace(' ', '_')].after = SSNO.AnyStandardName
                        else:
                            qualifications_dict[q_full_name.replace(' ', '_')].after = qualifications_dict[position - 1]

                # relate the phrases to each other
                # [component] standard_name [in_medium]
                # -1, 0, 1

        data_dict = {'title': data.get('name', data.get('title', None)),
                     'qualifiedAttribution': qualifiedAttribution,
                     'version': data.get('version', None),
                     'description': data.get('description', None),
                     'identifier': data.get('identifier', None),
                     'standardNames': [_parse_standard_names(k, v) for k, v in standardNames.items()]}
        if qualifications_dict:
            data_dict['hasModifier'] = list(qualifications_dict.values())
        if data.get('identifier', None):
            data_dict['id'] = data.get('identifier', None)

        return data_dict


class JSONLDReader(TableReader):
    def parse(self, **kwargs) -> Dict:
        from .standard_name_table import StandardNameTable
        with open(self.filename, 'r') as f:
            import json
            snt = StandardNameTable.from_jsonld(data=json.load(f), limit=1)
            return snt.model_dump(exclude_none=True)


_plugins = {
    'xml': XMLReader,
    'text/xml': XMLReader,
    'https://www.iana.org/assignments/media-types/text/xml': XMLReader,
    'https://www.iana.org/assignments/media-types/application/xml': XMLReader,
    'yaml': YAMLReader,
    'yml': YAMLReader,
    'application/yaml': YAMLReader,
    'https://www.iana.org/assignments/media-types/application/yaml': YAMLReader,
    'jsonld': JSONLDReader,
    'application/json-ld': JSONLDReader,
    'https://www.iana.org/assignments/media-types/application/ld+json': JSONLDReader
}


def get(plugin_name: str, default=None) -> Union[TableReader, None]:
    """Returns the plugin"""
    plugin = _plugins.get(str(plugin_name), None)
    if plugin is None:
        return default
    return plugin



import pathlib

import ssnolib
import utils
from ssnolib import prov, StandardNameTable

__this_dir__ = pathlib.Path(__file__).parent

from ssnolib.schema import Project, ResearchProject

CACHE_DIR = ssnolib.utils.get_cache_dir()


class TestSchema(utils.ClassTest):

    def test_project(self):
        snt_dist = ssnolib.dcat.Distribution(
            downloadURL="https://sandbox.zenodo.org/records/123202/files/Standard_Name_Table_for_the_Property_Descriptions_of_Centrifugal_Fans.jsonld",
            media_type="application/json+ld"
        )

        dataset = ssnolib.dcat.Dataset(
            identifier="https://sandbox.zenodo.org/uploads/123202",
            distribution=snt_dist
        )

        proj = Project(
            identifier='https://example.com/project',
            funder=prov.Organization(name='Funder'),
            usesStandardnameTable=dataset
        )

        self.assertEqual(str(proj.identifier), 'https://example.com/project')
        self.assertIsInstance(proj.funder, prov.Organization)
        self.assertEqual(proj.funder.name, 'Funder')
        self.assertEqual(proj.id, 'https://example.com/project')

    def test_research_project(self):
        snt = StandardNameTable(title="My SNT")
        proj = ResearchProject(
            identifier='https://example.com/research_project',
            funder=prov.Organization(name='Funder'),
            usesStandardnameTable=snt
        )
        self.assertEqual(proj.usesStandardnameTable.title, 'My SNT')
        self.assertEqual(proj.id, 'https://example.com/research_project')




import unittest

from ssnolib.ssno.unit_utils import _parse_unit, _get_ureg


class TestUnitParsing(unittest.TestCase):

    def test_malformed_unit(self):
        self.assertEqual(
            _parse_unit("m2/s3"),
            _get_ureg()("m**2/s**3").u
        )

    def test_malformed_unit2(self):
        self.assertEqual(
            _parse_unit("kg m-1 s-1"),
            _get_ureg()("kg/m/s").u
        )
